# src: https://github.com/facebookresearch/DrQA/blob/master/drqa/reader/data.py
import numpy as np

from torch.utils.data import Dataset
from torch.utils.data.sampler import Sampler
from .vector import vectorize


# ------------------------------------------------------------------------------
# PyTorch dataset class for MSMARCO data.
# ------------------------------------------------------------------------------


class RankerDataset(Dataset):
    def __init__(self, examples, model, shuffle=False):
        self.model = model
        self.examples = examples
        self.shuffle = shuffle

    def __len__(self):
        return len(self.examples)

    def __getitem__(self, index):
        return vectorize(self.examples[index],
                         self.model,
                         shuffle=self.shuffle)

    def lengths(self):
        return [(max([len(doc.tokens) for doc in ex.documents]),
                 len(ex.tokens)) for ex in self.examples]


# ------------------------------------------------------------------------------
# PyTorch sampler returning batched of sorted lengths (by doc and query).
# ------------------------------------------------------------------------------

class SortedBatchSampler(Sampler):
    def __init__(self, lengths, batch_size, shuffle=True):
        self.lengths = lengths
        self.batch_size = batch_size
        self.shuffle = shuffle

    def __iter__(self):
        lengths = np.array(
            [(-l[0], -l[1], np.random.random()) for l in self.lengths],
            dtype=[('l1', np.int_), ('l2', np.int_), ('rand', np.float_)]
        )
        indices = np.argsort(lengths, order=('l1', 'l2', 'rand'))
        batches = [indices[i:i + self.batch_size]
                   for i in range(0, len(indices), self.batch_size)]
        if self.shuffle:
            np.random.shuffle(batches)
        return iter([i for batch in batches for i in batch])

    def __len__(self):
        return len(self.lengths)
